-------------------------------------------------------------------------------
--! @file       CipherCore_Control.vhd
--! @author     Ekawat (ice) Homsirikamol
--! @brief      ASCON CipherCore
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.ALL;
use work.AEAD_pkg.all;

entity CipherCore is
    generic (
        --! Reset behavior
        G_ASYNC_RSTN    : boolean := False; --! Async active low reset
        --! Block size (bits)
        G_DBLK_SIZE     : integer := 64;    --! Data
        G_KEY_SIZE      : integer := 32;    --! Key
        G_TAG_SIZE      : integer := 128;   --! Tag
        --! The number of bits required to hold block size expressed in
        --! bytes = log2_ceil(G_DBLK_SIZE/8)
        G_LBS_BYTES      : integer := 4;
        --! Maximum supported AD/message/ciphertext length = 2^G_MAX_LEN-1
        G_MAX_LEN       : integer := SINGLE_PASS_MAX
    );
    port (
        --! Global
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! PreProcessor (data)
        key             : in  std_logic_vector(G_KEY_SIZE       -1 downto 0);
        bdi             : in  std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        --! PreProcessor (controls)
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;
        decrypt         : in  std_logic;
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(3                -1 downto 0);
        bdi_partial     : in  std_logic;
        bdi_eot         : in  std_logic;
        bdi_eoi         : in  std_logic;
        bdi_size        : in  std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        bdi_valid_bytes : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        bdi_pad_loc     : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        --! PostProcessor
        bdo             : out std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        bdo_valid       : out std_logic;
        bdo_ready       : in  std_logic;
        bdo_size        : out std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        msg_auth_done   : out std_logic;
        msg_auth_valid  : out std_logic
    );
end entity CipherCore;


architecture structure of CipherCore is
    signal is_first_dat : std_logic;
    signal en_key       : std_logic;
    signal en_state     : std_logic;
    signal en_npub      : std_logic;
    signal en_cmp       : std_logic;
    signal clr_rc       : std_logic;
    signal en_rc        : std_logic;
    signal sel_key_hi   : std_logic;
    signal sel_key_lo   : std_logic;
    signal sel_decrypt  : std_logic;
    signal sel_state    : std_logic_vector(2               -1 downto 0);
    signal sel_tag      : std_logic_vector(2               -1 downto 0);
begin
    u_cc_dp:
    entity work.CipherCore_Datapath(structure)
    port map (
        clk                 => clk              ,
        rst                 => rst              ,

        --! Input Processor
        key                 => key              ,
        bdi                 => bdi              ,
        bdi_valid_bytes     => bdi_valid_bytes  ,

        --! Datapath
        en_key              => en_key           ,
        en_state            => en_state         ,
        en_npub             => en_npub          ,
        en_cmp              => en_cmp           ,
        is_first_dat        => is_first_dat     ,
        clr_rc              => clr_rc           ,
        en_rc               => en_rc            ,
        sel_key_hi          => sel_key_hi       ,
        sel_key_lo          => sel_key_lo       ,
        sel_decrypt         => sel_decrypt      ,
        sel_state           => sel_state        ,
        sel_tag             => sel_tag          ,

        --! Output Processor
        bdo                 => bdo              ,
        msg_auth_valid      => msg_auth_valid
    );

    u_cc_ctrl:
    entity work.CipherCore_Control(behavior)
    port map (
        clk                 => clk              ,
        rst                 => rst              ,

        --! Input
        bdi                 => bdi              ,
        key_ready           => key_ready        ,
        key_valid           => key_valid        ,
        key_update          => key_update       ,
        decrypt             => decrypt          ,
        bdi_ready           => bdi_ready        ,
        bdi_valid           => bdi_valid        ,
        bdi_type            => bdi_type         ,
        bdi_eot             => bdi_eot          ,
        bdi_eoi             => bdi_eoi          ,
        bdi_size            => bdi_size         ,

        --! Datapath
        en_key              => en_key           ,
        en_state            => en_state         ,
        en_npub             => en_npub          ,
        en_cmp              => en_cmp           ,
        is_first_dat        => is_first_dat     ,
        clr_rc              => clr_rc           ,
        en_rc               => en_rc            ,
        sel_key_hi          => sel_key_hi       ,
        sel_key_lo          => sel_key_lo       ,
        sel_decrypt         => sel_decrypt      ,
        sel_state           => sel_state        ,
        sel_tag             => sel_tag          ,

        --! Output
        msg_auth_done       => msg_auth_done    ,
        bdo_ready           => bdo_ready        ,
        bdo_valid           => bdo_valid
    );
end structure;